// Unreal Fighting Framework by Kevin Poretti

#include "FFStateMachineComponent.h"

// FF includes
#include "FFState.h"

UFFStateMachineComponent::UFFStateMachineComponent()
{
    // Don't use Unreal's tick instead use a fixed tick
	PrimaryComponentTick.bCanEverTick = false;
}


void UFFStateMachineComponent::Initialize()
{
    UFFState* EntryStateInstance = AddState(EntryState);

    for(const TSubclassOf<UFFState>& CurrState : DefaultStates)
    {
        AddState(CurrState);
    }

    // need an entry state or something went seriously wrong
    check(EntryStateInstance);

    CurrentState = EntryStateInstance;
    CurrentState->Enter(GetCurrentStateContext());
}


void UFFStateMachineComponent::InitActorInfo(AActor* InOwner, AActor* InAvatar)
{
    Owner = InOwner;
	Avatar = InAvatar;
}


UFFState* UFFStateMachineComponent::AddState(TSubclassOf<UFFState> StateClassToAdd)
{
    UFFState* TempState = NewObject<UFFState>(this, StateClassToAdd);
    if(TempState)
    {
        if(!FindStateWithName(TempState->Name))
        {
            States.Add(TempState);
            return TempState;
        }

        UE_LOG(LogTemp, Error, TEXT("State with name %s already exists in state machine for %s"), 
            *TempState->Name.ToString(), *GetOwner()->GetFName().ToString());
        return nullptr;
    }

    UE_LOG(LogTemp, Error, TEXT("Could not create instance of state class %s"), 
        *StateClassToAdd.GetDefaultObject()->Name.ToString());

    return nullptr;
}


void UFFStateMachineComponent::AddStates(const TArray<TSubclassOf<UFFState>>& StateClassesToAdd)
{
    for(const TSubclassOf<UFFState>& CurrState : StateClassesToAdd)
    {
        AddState(CurrState);
    }
}


void UFFStateMachineComponent::RemoveState(FName StateToRemove)
{
    UE_LOG(LogTemp, Error, TEXT("UFFStateMachineComponent::RemoveState is not yet implemented"));
}

void UFFStateMachineComponent::SwitchStates(UFFState* NewState)
{
    check(CurrentState);
    check(NewState);

    CurrentState->Exit(GetCurrentStateContext());
    CurrentState = NewState;
    CurrentState->Enter(GetCurrentStateContext());
    TicksInState = 0;
    CurrentSubStateLabel = NAME_None;
}

void UFFStateMachineComponent::SwitchToEntryState()
{
    // can't have an entry state if there are no states
    check(States.Num() > 0);

    SwitchStates(States[0]);
}


int64 UFFStateMachineComponent::GetTicksInState() const
{
    return TicksInState;
}


FName UFFStateMachineComponent::GetCurrentStateName() const
{
    check(CurrentState)

    return CurrentState->Name;
}


FName UFFStateMachineComponent::GetCurrentSubStateLabel() const
{
    return CurrentSubStateLabel;
}


void UFFStateMachineComponent::SetSubStateLabel(FName InSubStateLabel)
{
    CurrentSubStateLabel = InSubStateLabel;
}


FFFStateContext UFFStateMachineComponent::GetCurrentStateContext()
{
    FFFStateContext CurrStateContext;
    CurrStateContext.Owner = Owner;
    CurrStateContext.Avatar = Avatar;
    CurrStateContext.Parent = this;
    return CurrStateContext;
}


void UFFStateMachineComponent::FixedTick(float OneFrame)
{
    // CurrentState should never be null
    // TODO: Should probably assert or whatever UE's equivalent is
    check(CurrentState);

    // Should we switch states?
    UFFState* StateToTransitionTo = nullptr;
    for(UFFState* ThisState : States)
    {
        // found a state 
        if(ThisState->CanTransition(GetCurrentStateContext()))
        {
            StateToTransitionTo = ThisState;
            break;
        }
    }

    // Lastly just check if the state we're about to transition into isn't the current state.
    // It is OK to transition if state's "CanTransitionToSelf" is true
    if(StateToTransitionTo && 
        (CurrentState->Name != StateToTransitionTo->Name || StateToTransitionTo->bCanTransitionToSelf))
    {
        SwitchStates(StateToTransitionTo);
    }
    else
    {
        // Tick current state
        TicksInState++;
        CurrentState->Update(OneFrame, GetCurrentStateContext());   
    }

    // Debug
}


void UFFStateMachineComponent::BeginPlay()
{
	Super::BeginPlay();

    Initialize();
}


UFFState* UFFStateMachineComponent::FindStateWithName(FName StateName)
{
    for (UFFState* CurrState : States)
    {
        if(CurrState->Name == StateName)
        {
            return CurrState;
        }
    }

	UE_LOG(LogTemp, Warning,
			TEXT("Could not find state in state machine with name %s on %s"), *StateName.ToString(), *Owner->GetName());

	return nullptr;
}